/*
 ****************************************************************************
 *
 * PC serial watchdog and power manager
 *
 ****************************************************************************
 */ 
#include "config.h"



/*
 ****************************************************************************
 *
 * CPU configuration
 *
 */
static __code unsigned short __at (CONFIG_ADDR) CONFIG = CONFIG_VALUE;

 

/*
 ****************************************************************************
 *
 * EEPROM configuration
 *
 */
static __code unsigned char __at (EEPROM_ADDR) EEPROM[] = {

	DEF_POR_TIMEOUT,
#define IDX_POR_TIMEOUT		0

	DEF_POWER_TIMEOUT,
#define IDX_POWER_TIMEOUT	1

	DEF_OFF_TIMEOUT,
#define IDX_OFF_TIMEOUT		2

	DEF_ON_TIMEOUT,
#define IDX_ON_TIMEOUT		3

	DEF_WD_TIMEOUT,
#define IDX_WD_TIMEOUT		4

	DEF_RESTART_TIMEOUT,
#define IDX_RESTART_TIMEOUT	5

	DEF_INT_TIMEOUT,
#define IDX_INT_TIMEOUT		6

	DEF_RESET_COUNT,
#define IDX_RESET_COUNT		7
};

#define MAX_EEPROM_IDX		7



/*
 ****************************************************************************
 *
 * Serial port command definitions. A command always consists of 9 bytes:
 *    0    -  Fixed value: 0x57
 *    1    -  Fixed value: 0x44
 *    2/3  -  Command byte (ascii hex)
 *    4/5  -  Argument value (ascii hex)
 *    6/7  -  Checksum (ascii hex)
 *    8    -  Fixed value: 0x0D
 *
 * The command bytes are as follows:
 *   0x01  -  No operation
 *   0x02  -  Turn watchdog on
 *   0x03  -  Turn watchdog off, also disables any reset timer
 *   0x04  -  Reset the system, the argument value gives the number of SYS_HZ
 *            cycles to wait
 *   0x05  -  Turn the system off, the argument value gives the number of
 *            SYS_HZ cycles to wait
 *   0x06  -  Return the watchdog status byte
 *   0x07  -  Return the firmware major version number
 *   0x08  -  Return the firmware minor version number
 *   0x09  -  Clear the watchdog reset counter
 *   0x1X  -  Return the specified EEPROM value
 *   0x2X  -  Set the specified EEPROM value
 *            store
 *
 */
#define CMD_LENGTH	9
#define CMD_ID1		0x57
#define CMD_ID2		0x44
#define CMD_END		0x0D

#define CMD_NOP		0x01
#define CMD_WD_ON	0x02
#define CMD_WD_OFF	0x03
#define CMD_RESET	0x04
#define CMD_SYS_OFF	0x05
#define CMD_STATUS	0x06
#define CMD_VER_MAJOR	0x07
#define CMD_VER_MINOR	0x08
#define CMD_CLR_RESET	0x09
#define CMD_GETVAL	0x10
#define CMD_SETVAL	0x20
#define CMD_INVALID	0x80	/* same as RESULT_NACK1 */

static unsigned char cmdbuf[3];
static unsigned char cmdlen;



/*
 ****************************************************************************
 *
 * Command results consist of 7 bytes:
 *   0    -  Fixed value: 0x57
 *   1    -  Fixed value: 0x44
 *   2/3  -  Command code or error code (ascii-hex)
 *   4/5  -  Result value (ascii-hex)
 *   6    -  Fixed value: 0x0D
 *
 * The result value is zero for commands not returning a specific value.
 *
 * The error code is one of:
 *   0x80  -  NACK1: Unknown command
 *   0x81  -  NACK2: Invalid command argument
 *   0x82  -  NACK3: Invalid checksum
 *   0x8F  -  INT:   Interrupt code
 *
 */
#define RESULT_LENGTH	7
#define RESULT_ID1	0x57
#define RESULT_ID2	0x44
#define RESULT_END	0x0D

#define RESULT_NACK1	0x80		/* invalid command */
#define RESULT_NACK2	0x81		/* invalid index */
#define RESULT_NACK3	0x82		/* invalid checksum */
#define RESULT_INT	0x8F

static unsigned char resultbuf[2];
static unsigned char resultlen;



/*
 ****************************************************************************
 *
 * Watchdog timeout value. This must be in RAM to be accessible from the
 * interrupt service routine.
 *
 */
static unsigned char wd_timeout;



/*
 ****************************************************************************
 *
 * Watchdog reset counter
 *
 */
static unsigned char wd_count;



/*
 ****************************************************************************
 *
 * System state
 *
 */
static union {
	unsigned char val;
	struct {
		unsigned char wd:1;		/* state of watchdog */
		unsigned char system:1;		/* state of system power */
		unsigned char dtr:1;		/* state of DTR line */
		unsigned char por:1;		/* power-on reset */
		unsigned char pwrsw:1;		/* power switch activated */
		unsigned char reset:1;		/* reset activated */
		unsigned char poweroff:1;	/* poweroff activated */
		unsigned char keyint:1;		/* interrupt key activated */
	};
} state;



/*
 ****************************************************************************
 *
 * Various system timers
 *
 */
static volatile unsigned char gen_timer;	/* general use timer */
static volatile unsigned char switch_timer;	/* switch timer */
static volatile unsigned char wd_timer;		/* watchdog timer */
static volatile unsigned char reset_timer;	/* reset timer */



/*
 ****************************************************************************
 *
 * Interrupt status
 *
 */
static volatile union {
	unsigned char val;
	struct {
		unsigned char swoldval:1;	/* old power switch value */
		unsigned char swactive:1;	/* switch press detected */
		unsigned char swinactive:1;	/* switch release detected */
		unsigned char received:1;	/* serial byte received */
		unsigned char rxactive:1;	/* receiving in progress */
		unsigned char txactive:1;	/* transmit in progress */
	};
} intstatus;



/*
 ****************************************************************************
 *
 * Serial port variables
 *
 */
static volatile unsigned char rxdata;		/* received data byte */
static volatile unsigned char rxreg;		/* receive register */
static volatile unsigned char txreg;		/* transmit register */
static volatile unsigned char bitcount;		/* current bit count */



/*
 ****************************************************************************
 *
 * Interrupt vector
 *
 */
void isr(void) __interrupt 0
{
  /*
   * Handle system timer interrupt
   */
  if (TMR1IF) {
	TMR1ON = 0;
	TMR1H = (TMR1_RELOAD >> 8) & 0xFF;
	TMR1L = TMR1_RELOAD & 0xFF;
	TMR1ON = 1;
	TMR1IF = 0;
	if (gen_timer > 0)
		gen_timer--;
	if (switch_timer > 0)
		switch_timer--;
	if (reset_timer > 0)
		reset_timer--;
	if (wd_timer > 0)
		wd_timer--;
  }


  /*
   * Handle serial port timer interrupt
   */
  if (T0IF) {
	TMR0 = TMR0_RELOAD_BIT;
	if (intstatus.txactive) {
		/* Transmit the next bit */
		if (bitcount > (STOP_BITS + IDLE_BITS)) {
			/* Send data bits */
			bitcount--;
			if (txreg & 0x01)
				TX_PORT = TX_HIGH;
			else
				TX_PORT = TX_LOW;
			txreg >>= 1;
		} else if (bitcount > 0) {
			/* Send stop and idle bits */
			bitcount--;
			TX_PORT = TX_HIGH;
		} else {
			/* Terminate transmission */
			intstatus.txactive = 0;
			T0IE = 0;
		}
	} else if (intstatus.rxactive) {
		/* Receive the next bit */
		if (bitcount > STOP_BITS) {
			/* Shift bit into receive register */
			bitcount--;
			rxreg >>= 1;
			if (!RX_PORT)
				rxreg |= 0x80;
		} else if (bitcount > 0) {
			/* Wait for stop bits */
			bitcount--;
		} else {
			/* Re-enable serial port for next byte */
			intstatus.rxactive = 0;
			if (!intstatus.received) {
				rxdata = rxreg;
				intstatus.received = 1;
			}
			T0IE = 0;
			RX_IF = 0;
			RX_IE = 1;
		}
	}
	T0IF = 0;
  }


  /*
   * Handle RX input interrupt. This starts a serial reception and also
   * resets the watchdog timer. In order to avoid glitches, we read the
   * receive input pin again: for a start bit it has to be set.
   */
  if (RX_IF) {
	if (RX_PORT) {
		wd_timer = wd_timeout;
		if (!intstatus.rxactive && !intstatus.txactive) {
			RX_IE = 0;
			TMR0 = TMR0_RELOAD_SBIT;
			T0IF = 0;
			T0IE = 1;
			rxreg = 0;
			bitcount = DATA_BITS + STOP_BITS;
			intstatus.rxactive = 1;
		}
	}
	RX_IF = 0;
  }


  /*
   * Handle int-on-change interrupt for power switch
   */
  if (GPIF) {
	if (SWITCH_PORT) {
		/* Low-to-high transition */
		intstatus.swoldval = 1;
		intstatus.swactive = 0;
		intstatus.swinactive = 1;
	} else {
		/* High-to-low transition */
		intstatus.swoldval = 0;
		intstatus.swactive = 1;
		intstatus.swinactive = 0;
	}
	GPIF = 0;
  }
}



/*
 ****************************************************************************
 *
 * Read a value from EEPROM
 *
 */
static unsigned char eeread(unsigned char idx)
{
  EEADR = idx;
  RD = 1;
  return(EEDATA);
}



/*
 ****************************************************************************
 *
 * Write a value into EEPROM
 *
 */
static void eewrite(unsigned char idx, unsigned char val)
{
  /* Prepare for write */
  EEADR = idx;
  EEDATA = val;
  WREN = 1;

  /* Initiate write */
  GIE = 0;
  EECON2 = 0x55;
  EECON2 = 0xAA;
  WR = 1;
  GIE = 1;

  /* Wait until finished */
  while (WR) ;
  WREN = 0;
}



/*
 ****************************************************************************
 *
 * Turn the watchdog and reset timers off
 *
 */
static void wd_off(void)
{
  if (state.wd) {
	state.wd = 0;
	wd_timer = 0;
	wd_timeout = 0;
  }

  if (state.reset || state.poweroff) {
	state.reset = 0;
	state.poweroff = 0;
	reset_timer = 0;
  }
}



/*
 ****************************************************************************
 *
 * Turn the watchdog on
 *
 */
static void wd_on(void)
{
  if (!state.por && !state.wd && state.system) {
	state.wd = 1;
	wd_timeout = eeread(IDX_WD_TIMEOUT);
	wd_timer = wd_timeout;
  }
}



/*
 ****************************************************************************
 *
 * Activate system power switch. This also disables the watchdog timer.
 * We don't do anything else during this time.
 *
 */
static void toggle_power(void)
{
  /* Toggle the power port for some time */
  POWER_PORT = 0;
  gen_timer = eeread(IDX_POWER_TIMEOUT);
  while (gen_timer > 0) ;
  POWER_PORT = 1;
}



/*
 ****************************************************************************
 *
 * Set voltage reference value
 *
 */
static void set_vref(unsigned char val)
{
  unsigned char i;

  VRCON = val;
  for (i = 0; i < 4; i++)
	__asm nop __endasm;
}



/*
 ****************************************************************************
 *
 * Determine the system status
 *
 */
static void get_status(void)
{
  /*
   * If the voltage on the DTR line is below -3V, the system is powered, but
   * the serial pofrt has been turned off. If the voltage is above +3V, the system
   * is powered and the serial port is turned on, and if the voltage is in between
   * the system is powered off.
   */
  set_vref(DTR_VR_LOW);
  if (DTR_PORT) {
	/* System is on, but serial port is turned off */
	state.system = 1;
	state.dtr = 0;
  } else {
	set_vref(DTR_VR_HIGH);
	if (!DTR_PORT) {
		/* System is on, and serial port is turned on */
		state.system = 1;
		state.dtr = 1;
	} else {
		set_vref(DTR_VR_LOW);
		if (DTR_PORT) {
			/* The DTR line changed between the checks */
			state.system = 1;
			state.dtr = 0;
		} else {
			/* System is off, and the serial port has to be off as well */
			state.system = 0;
			state.dtr = 0;
		}
	}
  }
}



/*
 ****************************************************************************
 *
 * Convert value in result buffer into hex ascii
 *
 */
static unsigned char puthex(unsigned char idx)
{
  unsigned char val;

  val = resultbuf[idx];
  resultbuf[idx] <<= 4;
  val = (val >> 4) + '0';
  if (val > '9')
	val += 'A' - '9' - 1;
  return(val);
}



/*
 ****************************************************************************
 *
 * Convert hex ascii value into binary value
 *
 */
static unsigned char gethex(unsigned char idx, unsigned char val)
{
  if (val < '0')
	return(FALSE);
  else if (val <= '9')
	val -= '0';
  else if (val < 'A' || val > 'F')
	return(FALSE);
  else
	val -= 'A' - 10;
  cmdbuf[idx] = (cmdbuf[idx] << 4) + val;
  return(TRUE);
}



/*
 ****************************************************************************
 *
 * Disable the serial port receiver
 *
 */
static void rx_disable(void)
{
  RX_IE = 0;
  intstatus.rxactive = 0;
  intstatus.received = 0;
  if (!intstatus.txactive)
	T0IE = 0;
}



/*
 ****************************************************************************
 *
 * Enable the serial port receiver
 *
 */
static void rx_enable(void)
{
  if (!RX_IE && !intstatus.rxactive && !intstatus.txactive) {
	T0IE = 0;
	RX_IF = 0;
	RX_IE = 1;
  }
}



/*
 ****************************************************************************
 *
 * Disable the serial port transmitter
 *
 */
static void tx_disable(void)
{
  if (intstatus.txactive) {
	intstatus.txactive = 0;
	TX_PORT = TX_HIGH;
	T0IE = 0;
  }
}



/*
 ****************************************************************************
 *
 * Enable the serial port transmitter and start sending a byte
 *
 */
static void tx_enable(unsigned char c)
{
  if (!intstatus.txactive) {
	T0IE = 0;
	txreg = c;
	intstatus.txactive = 1;
	bitcount = DATA_BITS + STOP_BITS + IDLE_BITS;
	TX_PORT = TX_LOW;
	TMR0 = TMR0_RELOAD_BIT;
	T0IF = 0;
	T0IE = 1;
  }
}



/*
 ****************************************************************************
 *
 * Handle the serial port command interface
 *
 */
static void do_cmd(void)
{
  unsigned char val, res;

  /* Disable serial port if it is not activated */
  if (!state.dtr) {
	rx_disable();
	tx_disable();
	resultlen = 0;
	return;
  }

  /* Wait for the transmitter to become ready before sending or receiving */
  if (intstatus.txactive)
	return;

  /*
   * Check for sending an interrupt message. Delay the interrupt message
   * if we are currently receiving a command or sending a result reply.
   */
  if (state.keyint && cmdlen == 0 && resultlen == 0) {
	state.keyint = 0;
	resultlen = RESULT_LENGTH;
	resultbuf[0] = RESULT_INT;
	resultbuf[1] = 0;
  }

  /* Send any previous result */
  if (resultlen > 0) {

	/* Disable the receiver */
	rx_disable();

	/* Determine the next character to send */
	val = '*';
	switch (resultlen) {
		case (RESULT_LENGTH - 0):
			val = RESULT_ID1;
			break;
		case (RESULT_LENGTH - 1):
			val = RESULT_ID2;
			break;
		case (RESULT_LENGTH - 2):
		case (RESULT_LENGTH - 3):
			val = puthex(0);
			break;
		case (RESULT_LENGTH - 4):
		case (RESULT_LENGTH - 5):
			val = puthex(1);
			break;
		case (RESULT_LENGTH - 6):
			val = RESULT_END;
			break;
	}
	resultlen--;

	/* Send the next character */
	tx_enable(val);
	return;
  }

  /* Check if we have a new byte in the receive buffer */
  rx_enable();
  if (!intstatus.received)
	return;
  val = rxdata;
  intstatus.received = 0;

  /* Check the next command byte from the serial port */
  if (val != CMD_END) {
	switch (cmdlen) {
		case 0:
			if (val == CMD_ID1)
				cmdlen++;
			else if (val != 0x0A)
				cmdlen = CMD_LENGTH;
			break;

		case 1:
			if (val == CMD_ID2)
				cmdlen++;
			else
				cmdlen = CMD_LENGTH;
			break;

		case 2:
		case 3:
			if (gethex(0, val))
				cmdlen++;
			else
				cmdlen = CMD_LENGTH;
			break;

		case 4:
		case 5:
			if (gethex(1, val))
				cmdlen++;
			else
				cmdlen = CMD_LENGTH;
			break;

		case 6:
		case 7:
			if (gethex(2, val))
				cmdlen++;
			else
				cmdlen = CMD_LENGTH;
			break;
	}
	return;
  }

  /* When receiving a command end character, turn off the receiver */
  rx_disable();

  /* Check if we received a valid command string */
  if (cmdlen == 0) {
	cmdbuf[0] = CMD_NOP;
	cmdbuf[2] = 0;
  } else if (cmdlen != (CMD_LENGTH - 1)) {
	cmdbuf[0] = CMD_INVALID;
	cmdbuf[2] = 0;
  }
  cmdlen = 0;

  /* Process the command */
  res = cmdbuf[0];
  val = 0;
  if ((cmdbuf[2] != 0) &&
      (unsigned char)((cmdbuf[0] + cmdbuf[1] + cmdbuf[2]) & 0xFF != 0)) {
	/* Invalid checksum */
	res = RESULT_NACK3;
	val = cmdbuf[2];
  } else if ((cmdbuf[0] & 0xF0) == CMD_GETVAL) {
	/* Get an EEPROM value */
	val = cmdbuf[0] & 0x0F;
	if (val > MAX_EEPROM_IDX)
		res = RESULT_NACK2;
	else
		val = eeread(val);
  } else if ((cmdbuf[0] & 0xF0) == CMD_SETVAL) {
	/* Set an EEPROM value */
	val = cmdbuf[0] & 0x0F;
	if (val > MAX_EEPROM_IDX)
		res = RESULT_NACK2;
	else
		eewrite(val, cmdbuf[1]);
  } else switch (cmdbuf[0]) {
	case CMD_NOP:
		break;
	case CMD_WD_ON:
		wd_on();
		break;
	case CMD_WD_OFF:
		wd_off();
		break;
	case CMD_RESET:
		state.reset = 1;
		reset_timer = cmdbuf[1];
		break;
	case CMD_SYS_OFF:
		state.poweroff = 1;
		reset_timer = cmdbuf[1];
		break;
	case CMD_STATUS:
		val = state.val;
		break;
	case CMD_VER_MAJOR:
		val = VER_MAJOR;
		break;
	case CMD_VER_MINOR:
		val = VER_MINOR;
		break;
	case CMD_CLR_RESET:
		wd_count = 0;
		break;
	default:
		res = RESULT_NACK1;
		break;
  }

  /* Setup the result */
  resultlen = RESULT_LENGTH;
  resultbuf[0] = res;
  resultbuf[1] = val;
}



/*
 ****************************************************************************
 *
 * Check for interrupt key status. This routine gets called upon power
 * key release.
 *
 */
static void check_int(void)
{
  unsigned char i, j;

  /* If system not powered on, there can be no interrupt key activation */
  if (!state.system || !state.dtr) {
	state.keyint = 0;
	return;
  }

  /* Determine the relevant timeout values */
  i = eeread(IDX_INT_TIMEOUT);
  j = eeread(IDX_OFF_TIMEOUT);
  if (i == 0 || i > (j - 1)) {
	state.keyint = 0;
	return;
  }
  j -= i;

  /* Check if the timer has reached the timeout value */
  i = switch_timer;
  if (i > 0 && i < j)
	state.keyint = 1;
  else
	state.keyint = 0;
}



/*
 ****************************************************************************
 *
 * Main program
 *
 */
void main(void)
{
  /* Initially disable all interrupts */
  INTCON = 0;
  PIE1 = 0;
  IOC = 0;

  /* Check type of startup */
  if (!NOT_POR) {
	/* Power-on reset */
	NOT_POR = 1;
	state.val = 0;
	state.por = 1;
  } else {
	/* All other resets - we assume the system state not to have changed */
	state.por = 0;
  }

  /* Initialize I/O ports */
  CMCON = 0x07;
  GPIO = 0xFF;
  TRISIO = IO_TRIS;
  WPU = IO_WPU;
  OPTION_REG = 0x0F | RX_EDGE;

  /* Setup watchdog as disabled */
  state.wd = 0;
  state.reset = 0;
  state.poweroff = 0;
  wd_count = 0;
  wd_timer = 0;
  reset_timer = 0;

  /* Enable master interrupt flag. Individual interrupts are still disabled */
  PEIE = 1;
  GIE = 1;

  /* Initialize the power switch input */
  intstatus.swoldval = SWITCH_PORT;
  intstatus.swactive = 0;
  intstatus.swinactive = 0;
  state.pwrsw = 0;
  state.keyint = 0;
  switch_timer = 0;
  SWITCH_IOC = 1;
  GPIF = 0;
  GPIE = 1;

  /* Initialize the serial port */
  intstatus.received = 0;
  intstatus.rxactive = 0;
  intstatus.txactive = 0;
  state.dtr = 0;
  VRCON = DTR_VR_LOW;
  CMCON = DTR_CMCON;
  TX_PORT = TX_HIGH;
#if TMR0_PRESCALE == 0
  OPTION_REG = 0x0F | RX_EDGE;
#else
  OPTION_REG = 0x00 | (TMR0_PRESCALE - 1) | RX_EDGE;
#endif

  /* Initialize the command handler */
  cmdlen = 0;
  resultlen = 0;

  /* Initialize the system timer */
  T1CON = TMR1_PRESCALE << 4;
  TMR1H = (TMR1_RELOAD >> 8) & 0xFF;
  TMR1L = TMR1_RELOAD & 0xFF;
  TMR1IF = 0;
  TMR1IE = 1;
  TMR1ON = 1;

  /* After POR wait to enable system power - don't do anything else */
  if (state.por) {
	gen_timer = eeread(IDX_POR_TIMEOUT);
	while (!state.system) {
		get_status();
		if (gen_timer == 0) {
			toggle_power();		/* modifies gen_timer!! */
			break;
		}
	}
	state.por = 0;
  }

  /* After the system has been turned on, keep it running for some time */
  gen_timer = eeread(IDX_RESTART_TIMEOUT);
  while (gen_timer > 0) ;

  /* Main program loop */
  while(1) { 

	/*
	 * Determine the system status according to the voltage of the DTR
	 * input pin.
	 */
	get_status();

	/*
	 * Handle any serial port command.
	 */
	do_cmd();

	/*
	 * Check if the power-on switch has been activated. It needs to get
	 * pressed for at least four seconds until we actually switch the
	 * power.
	 */
	if (intstatus.swactive) {
		/* Activate switch timer on key press */
		intstatus.swactive = 0;
		switch_timer = eeread(state.system ? IDX_OFF_TIMEOUT : IDX_ON_TIMEOUT);
		state.pwrsw = 1;
		state.keyint = 0;
	} else if (intstatus.swinactive) {
		/* Key has been released before timeout */
		intstatus.swinactive = 0;
		state.pwrsw = 0;
		check_int();
	} else if (state.pwrsw && switch_timer == 0) {
		/* Toggle power if switch timer expired */
		toggle_power();
		wd_off();
		wd_count = 0;
		state.pwrsw = 0;
		state.keyint = 0;
	}

	/*
	 * Check if the watchdog or reset timers have expired. Only check this if
	 * the system is powered on.
	 */
	if (state.system &&
	    ((state.wd && wd_timer == 0) ||
	     ((state.reset || state.poweroff) && reset_timer == 0))) {
		unsigned char max_count = eeread(IDX_RESET_COUNT);

		/* Toggle or switch power if watchdog timer expired */
		toggle_power();
		if (state.reset ||
		    (state.wd && (max_count == 255 || wd_count < max_count))) {
			if (state.wd && max_count < 255)
				wd_count++;
			gen_timer = eeread(IDX_RESTART_TIMEOUT);
			while (gen_timer > 0) ;
			toggle_power();
		}
		wd_off();
		state.pwrsw = 0;
		state.keyint = 0;
	}
  }
}

