/*
 * watchdog.h  -  Definitions for hardware watchdog
 *
 * Copyright (C) 2008 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id$
 */

#ifndef WATCHDOG_H
#define WATCHDOG_H


/*
 ****************************************************************************
 *
 * Default serial device name
 *
 */
#ifndef PATH_SERIALPORT
# ifdef _PATH_DEV
#  define PATH_SERIALPORT	_PATH_DEV "serialwd"
# else
#  define PATH_SERIALPORT	"/dev/serialwd"
# endif
#endif



/*
 ****************************************************************************
 *
 * Default lock file path name
 *
 */
#ifndef PID_FILE
# ifdef _PATH_VARRUN
#  define PID_FILE	_PATH_VARRUN "watchdog.pid"
# else
#  ifdef _PATH_VARTMP
#   define PID_FILE	_PATH_VARTMP "watchdog.pid"
#  else
#   ifdef _PATH_TMP
#    define PID_FILE	_PATH_TMP "watchdog.pid"
#   else
#    define PID_FILE	"/tmp/watchdog.pid"
#   endif
#  endif
# endif
#endif



/*
 ****************************************************************************
 *
 * Serial port speed
 *
 */
#define WD_SPEED	1200



/*
 ****************************************************************************
 *
 * Watchdog timer tick in milliseconds
 *
 */
#define WD_TICK		100



/*
 ****************************************************************************
 *
 * Timeout between receiving watchdog characters in milliseconds. This
 * depends from the serial port speed and some voodoo.
 *
 */
#define WD_RCV_TIMEOUT		200



/*
 ****************************************************************************
 *
 * Timeout from sending a command until receiving the reply
 *
 */
#define WD_REPLY_TIMEOUT	500



/*
 ****************************************************************************
 *
 * Minimum and maximum watchdog timeout values in timer ticks
 *
 */
#define WD_TIMEOUT_MIN		50
#define WD_TIMEOUT_MAX		255



/*
 ****************************************************************************
 *
 * Configuration register index number range
 *
 */
#define WD_MIN_CONFIG	0
#define WD_MAX_CONFIG	7



/*
 ****************************************************************************
 *
 * Watchdog commands
 *
 */
#define WD_CMD_NOP		0x01
#define WD_CMD_WD_ON		0x02
#define WD_CMD_WD_OFF		0x03
#define WD_CMD_RESET		0x04
#define WD_CMD_SYS_OFF		0x05
#define WD_CMD_STATUS		0x06
#define WD_CMD_VER_MAJOR	0x07
#define WD_CMD_VER_MINOR	0x08
#define WD_CMD_CLR_RESET	0x09
#define WD_CMD_GETVAL		0x10
#define WD_CMD_SETVAL		0x20



/*
 ****************************************************************************
 *
 * Watchdog reply codes
 *
 */
#define WD_RESULT_NACK		0x80
#define WD_RESULT_NACK1		0x80		/* invalid command */
#define WD_RESULT_NACK2		0x81		/* invalid index */
#define WD_RESULT_NACK3		0x82		/* invalid checksum */
#define WD_RESULT_INT		0x8F



/*
 ****************************************************************************
 *
 * Error return codes. Since the normal return values can have at most a
 * value of 255, the error codes start above this.
 *
 */
#define ERR_NONE	256	/* No error */
#define ERR_SEND	257	/* Unable to send (errno) */
#define ERR_SELECT	258	/* Select error (errno) */
#define ERR_READ	259	/* Read error (errno) */
#define ERR_INVALID	260	/* Invalid response */
#define ERR_TO_REPLY	261	/* Timeout waiting for reply */
#define ERR_TO_CHAR	262	/* Timeout waiting for character */
#define ERR_NACK	263	/* Watchdog replied with NACK */
#define ERR_NACK1	(ERR_NACK + 0)
#define ERR_NACK2	(ERR_NACK + 1)
#define ERR_NACK3	(ERR_NACK + 2)



/*
 ****************************************************************************
 *
 * Number of retries when sending a watchdog command
 *
 */
#define WD_RETRIES	3



/*
 ****************************************************************************
 *
 *  Number of seconds to wait after killing the daemon
 *
 */
#define WAIT_KILL	20



/*
 ****************************************************************************
 *
 * External routines
 *
 */

/* Initialize log file handling */
extern void log_init(const char *fname);


/* Close the log file */
extern void log_close(void);


/* Write a message to the log file */
extern void log_message(bool iserror, const char *msg);


/* Write an error with errno to the log file */
extern void log_error(const char *msg);


/* Open the watchdog serial port */
extern void wd_open(const char *devname);


/* Close the watchdog serial port */
extern void wd_close(void);


/* Initialize the watchdog interface */
extern char *wd_init(void);


/* Get watchdog status */
extern unsigned int wd_status(void);


/* Get a watchdog configuration value */
extern unsigned int wd_getval(unsigned int idx);


/* Set a watchdog configuration value */
extern void wd_setval(unsigned int idx, unsigned int val);


/* Clear the watchdog reset counter */
extern void wd_resetclr(void);


/* Setup the watchdog timeout */
extern unsigned int wd_setup(void);


/* Turn the watchdog on */
extern bool wd_daemon_on(void);


/* Turn the watchdog off */
extern void wd_daemon_off(void);


/* Retrigger the watchdog at the next timer event */
extern bool wd_daemon_sleep(void);



/*
 ****************************************************************************
 *
 * Global variables
 *
 */

/* Name of current program */
extern char *progname;


/* Verbosity level */
extern int verbose;


#endif

