/*
 * wd.c  -  Serial watchdog I/O interface handling
 *
 * Copyright (C) 2008 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id$
 */


#include "compat.h"
#include "watchdog.h"
#include <ctype.h>
#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif
#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#else
# ifdef HAVE_SYS_POLL_H
#  include <sys/poll.h>
# endif
#endif



/*
 ****************************************************************************
 *
 * Serial port parameters
 *
 */
#if WD_SPEED == 300
# define BAUDRATE	B300
#elif WD_SPEED == 1200
# define BAUDRATE	B1200
#elif WD_SPEED == 2400
# define BAUDRATE	B2400
#elif WD_SPEED == 4800
# define BAUDRATE	B4800
#elif WD_SPEED == 9600
# define BAUDRATE	B9600
#else
# error Invalid serial port speed
#endif



/*
 ****************************************************************************
 *
 * Local variables
 *
 */

/* Serial port device */
static int serialfd = -1;


/* Old termios status */
static struct termios oldtio;


/* Watchdog timeout in ms */
static unsigned int wd_timeout = 0;


/* Flag if watchdog is active */
static bool wd_active = false;



/*
 ****************************************************************************
 *
 * Open the serial watchdog device
 *
 */
void wd_open(const char *devname)
{
  struct stat sb;
  struct termios tio;
  int fd;

  /* Check if the file name is a character device */
  if (stat(devname, &sb) < 0) {
	log_error("Unable to access watchdog serial port");
	exit(1);
  }
  if ((sb.st_mode & S_IFMT) != S_IFCHR) {
	log_error("Watchdog serial port file is not a character device");
	exit(1);
  }

  /* Open the serial port device */
  if ((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
	log_error("Unable to open watchdog serial port");
	exit(1);
  }

  /* Get current termios status */
  if (tcgetattr(fd, &oldtio) == -1) {
	log_error("Unable to get attributes of watchdog serial port");
	close(fd);
	exit(1);
  }

  /* Set the serial port to new parameters */
  memzero(&tio, sizeof(tio));
  tio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD | CSTOPB;
  tio.c_iflag = IGNBRK | IGNPAR;
  tio.c_oflag = 0;
  tio.c_lflag = 0;
  if (tcsetattr(fd, TCSANOW, &tio) == -1) {
	log_error("Unable to set attributes of watchdog serial port");
	close(fd);
	exit(1);
  }

  /* Set the file descriptor for future reference */
  serialfd = fd;
}



/*
 ****************************************************************************
 *
 * Close the serial watchdog device
 *
 */
void wd_close(void)
{
  if (serialfd != -1) {
	tcsetattr(serialfd, TCSANOW, &oldtio);
	close(serialfd);
	serialfd = -1;
  }
}



/*
 ****************************************************************************
 *
 * Handle an interrupt message. Currently, this does nothing more but just
 * logging.
 *
 */
static void handle_int(void)
{
  if (wd_active)
	log_message(false, "Received watchdog interrupt");
}



/*
 ****************************************************************************
 *
 * Print a send/receive buffer if debugging
 *
 */
static void print_debug(bool send, const char *buf)
{
  static char str1[40];
  char str2[20];
  int i, j;

  if (verbose > 1) {
	for (i = 0, j = 0; buf[i] != '\0'; i++) {
		if (isprint(buf[i]))
			str2[j++] = buf[i];
		else {
			str2[j++] = '\\';
			if (buf[i] == '\r')
				str2[j++] = 'r';
			else if (buf[i] == '\n')
				str2[j++] = 'n';
			else
				str2[j++] = '?';
		}
	}
	str2[j] = '\0';
	if (send)
		sprintf(str1, "Sending '%s' to watchdog", str2);
	else
		sprintf(str1, "Received '%s' from watchdog", str2);
	log_message(false, str1);
  }
}



/*
 ****************************************************************************
 *
 * Send a watchdog command
 *
 */
static bool send_cmd(unsigned int cmd, unsigned int arg)
{
  char buf[10];
  unsigned int cksum;
  size_t len;

  if (cmd == WD_CMD_NOP) {
	buf[0] = '\r';
	buf[1] = '\0';
	len = 1;
  } else {
	cksum = (256 - (cmd + arg)) & 0xFF;
	sprintf(buf, "WD%02X%02X%02X\r", cmd, arg, cksum);
	len = 9;
  }
  print_debug(true, buf);
  usleep(15000);
  tcflush(serialfd, TCIOFLUSH);
  if (write(serialfd, buf, len) != len)
	return(false);
  tcdrain(serialfd);
  return(true);
}



/*
 ****************************************************************************
 *
 * Wait for a command reply from watchdog. This routine returns false
 * if the read timed out.
 *
 */
static unsigned int wait_reply(unsigned int timeout,
				unsigned int *res, unsigned int *val)
{

#if !defined(HAVE_SELECT) && !defined(HAVE_POLL)
# error Must have either poll() or select()
#endif

  char buf[8];
#ifdef HAVE_SELECT
  struct timeval to;
  fd_set fds;
#else
  struct pollfd fds;
#endif
  size_t len;
  int idx, i;

#ifdef HAVE_SELECT
  /* Initialize the timeout */
  to.tv_usec = ((unsigned long)timeout * 1000UL) % 1000000UL;
  to.tv_sec = ((unsigned long)timeout * 1000UL) / 1000000UL;
#endif

  /* Wait for response */
  idx = 0;
  while (true) {
#ifdef HAVE_SELECT
	FD_ZERO(&fds);
	FD_SET(serialfd, &fds);
	i = select(serialfd + 1, &fds, NULL, NULL, &to);
#else
	fds.fd = serialfd;
	fds.events = POLLIN;
	fds.revents = 0;
	i = poll(&fds, 1, timeout);
#endif
	if (i == 0) {
		/* Timeout */
		return(idx == 0 ? ERR_TO_REPLY : ERR_TO_CHAR);
	} else if (i == -1) {
		/* Select/poll error */
		return(ERR_SELECT);
#ifdef HAVE_SELECT
	} else if (FD_ISSET(serialfd, &fds)) {
#else
	} else if ((fds.revents & POLLIN) != 0) {
#endif
		/* Got a character */
		len = read(serialfd, &(buf[idx]), 1);
		if (len == -1) {
			/* Read error */
			return(ERR_READ);
		} else if (len == 1) {
			if (buf[idx] == '\r')
				break;
			if (idx < 7)
				idx++;
		}
	}
#ifdef HAVE_SELECT
	to.tv_usec = (WD_RCV_TIMEOUT * 1000UL) % 1000000UL;
	to.tv_sec = (WD_RCV_TIMEOUT * 1000UL) / 1000000UL;
#else
	timeout = WD_RCV_TIMEOUT;
#endif
  }

  /* Check if we received an invalid reply */
  buf[idx + 1] = '\0';
  print_debug(false, buf);
  if (idx != 6 || sscanf(buf, "WD%2x%2x\r", res, val) != 2)
	return(ERR_INVALID);
  return(ERR_NONE);
}



/*
 ****************************************************************************
 *
 * Process a watchdog command. It returns the result value or an error
 * code. If an interrupt message arrives, we can assume that sending the
 * command was unsuccessful (the watchdog disables sending an interrupt
 * if it received a command), so we need to repeat sending the command.
 *
 */
static unsigned int do_command(unsigned int cmd, unsigned int arg)
{
  unsigned int res, val, err;
  unsigned int timeout, retry;
  bool gotint;

  /* Send the command and wait for an answer */
  retry = 0;
  gotint = false;
  while (true) {
	/* Send the command */
	if (!send_cmd(cmd, arg))
		return(ERR_SEND);

	/* Wait for an answer to our command */
	timeout = WD_REPLY_TIMEOUT;
	do {
		err = wait_reply(timeout, &res, &val);
		timeout = WD_RCV_TIMEOUT;
		if (err == ERR_NONE && res == WD_RESULT_INT) {
			/*
			 * In case of interrupt try reading again. If we
			 * don't get a timely answer, the interrupt causes
			 * a timeout error, which in turn causes a command
			 * resend lateron in this routine.
			 */
			gotint = true;
			res = 0;
		}
	} while (err == ERR_NONE && res < WD_RESULT_NACK && res != cmd);

	/* Check for response code */
	if (err == ERR_READ || err == ERR_SELECT || err == ERR_TO_CHAR) {
		/* Serious error, no need to repeat */
		return(err);
	} else if (err != ERR_NONE) {
		/* Reply timeout or invalid reply - try again */
		retry++;
		if (retry > WD_RETRIES)
			return(err);
	} else if (res < WD_RESULT_NACK) {
		/* Got a valid reply - break out */
		break;
	} else {
		/* Got a reply but with an error code - try again */
		retry++;
		if (retry > WD_RETRIES)
			return(res - WD_RESULT_NACK + ERR_NACK);
	}
  }

  /*
   * We got a valid answer, but before returning, handle any received
   * interrupt message.
   */
  if (gotint)
	handle_int();
  return(val);
}



/*
 ****************************************************************************
 *
 * Print error message and terminate
 *
 */
static bool wd_error(unsigned int err)
{
  /* Check if we have an error at all */
  if (err <= ERR_NONE)
	return(true);

  /* Print error message */
  switch (err) {
	case ERR_SEND:
		log_error("Unable to send to watchdog");
		break;
	case ERR_SELECT:
		log_error("Select error");
		break;
	case ERR_READ:
		log_error("Unable to read from watchdog");
		break;
	case ERR_INVALID:
		log_message(true, "Invalid response from watchdog");
		break;
	case ERR_TO_REPLY:
		log_message(true, "Timed out waiting for watchdog reply");
		break;
	case ERR_TO_CHAR:
		log_message(true, "Timed out reading from watchdog");
		break;
	case ERR_NACK1:
		log_message(true, "Invalid command sent to watchdog");
		break;
	case ERR_NACK2:
		log_message(true, "Invalid register index sent to watchdog");
		break;
	case ERR_NACK3:
		log_message(true, "Invalid checksum sent to watchdog");
		break;
	default:
		log_message(true, "Unknown error");
		break;
  }

  /* Exit if we are not called from the watchdog daemon */
  if (!wd_active) {
	wd_close();
	exit(1);
  }
  return(false);
}



/*
 ****************************************************************************
 *
 * Initialize the watchdog interface
 *
 */
char *wd_init(void)
{
  static char buf[10];
  unsigned int major, minor;

  /* Check if the watchdog answers */
  wd_error(do_command(WD_CMD_NOP, 0));

  /* Determine major version number */
  major = do_command(WD_CMD_VER_MAJOR, 0);
  wd_error(major);

  /* Determine minor version number */
  minor = do_command(WD_CMD_VER_MINOR, 0);
  wd_error(minor);

  /* Return version string */
  sprintf(buf, "%u.%u", major, minor);
  return(buf);
}



/*
 ****************************************************************************
 *
 * Print watchdog status
 *
 */
unsigned int wd_status(void)
{
  unsigned int i;

  i = do_command(WD_CMD_STATUS, 0);
  wd_error(i);
  return(i);
}



/*
 ****************************************************************************
 *
 * Print watchdog configuration value
 *
 */
unsigned int wd_getval(unsigned int idx)
{
  unsigned int i;

  i = do_command(WD_CMD_GETVAL + idx, 0);
  wd_error(i);
  return(i);
}



/*
 ****************************************************************************
 *
 * Set watchdog configuration value
 *
 */
void wd_setval(unsigned int idx, unsigned int val)
{
  wd_error(do_command(WD_CMD_SETVAL + idx, val));
}



/*
 ****************************************************************************
 *
 * Clear the watchdog reset counter
 *
 */
void wd_resetclr(void)
{
  wd_error(do_command(WD_CMD_CLR_RESET, 0));
}



/*
 ****************************************************************************
 *
 * Setup the watchdog timeout. The timeout must leave at least 2 seconds
 * before the hardware actually starts to reset the system. This is for
 * processing within wd_daemon_sleep().
 *
 */
unsigned int wd_setup(void)
{
  unsigned int i;

  /* Determine the watchdog timeout value */
  i = wd_getval(4);
#if WD_TIMEOUT_MAX < 255
  if (i > WD_TIMEOUT_MAX) {
	/* Limit the timeout value */
	i = WD_TIMEOUT_MAX;
  } else
#endif
  if (i == 0) {
	log_message(true, "Watchdog deactivated, check timeout value");
	wd_close();
	exit(1);
  } else if (i < WD_TIMEOUT_MIN) {
	log_message(true, "Watchdog timeout value %u too small");
	wd_close();
	exit(1);
  }
  wd_timeout = (i * WD_TICK * 2) / 3;
  if (wd_timeout > (i * WD_TICK - 2000))
	wd_timeout = i * WD_TICK - 2000;
  return(wd_timeout);
}



/*
 ****************************************************************************
 *
 * Turn the watchdog on. This routine needs to get called in the context
 * of the daemon process.
 *
 */
bool wd_daemon_on(void)
{
  wd_active = true;
  return(wd_error(do_command(WD_CMD_WD_ON, 0)));
}



/*
 ****************************************************************************
 *
 * Turn the watchdog off. This routine needs to get called in the context
 * of the daemon process. It sends the turn-off-command without waiting
 * for a response.
 *
 */
void wd_daemon_off(void)
{
  if (!wd_error(do_command(WD_CMD_WD_OFF, 0))) {
	/* Force shutdown again */
	do_command(WD_CMD_WD_OFF, 0);
  }
  wd_close();
  wd_active = false;
}



/*
 ****************************************************************************
 *
 * Wait for the next watchdog timer event and retrigger the watchdog
 *
 */
bool wd_daemon_sleep(void)
{
  unsigned int res, val, err, retry;

  /* Sleep a while before retriggering the watchdog */
  err = wait_reply(wd_timeout, &res, &val);
  if (err == ERR_NONE) {
	/* We received something, ignore anything but an interrupt */
	if (res == WD_RESULT_INT)
		handle_int();
  } else if (err == ERR_INVALID) {
	/* We received an invalid message, sleep a little bit and send a NOP */
	sleep(1);
  } else if (err != ERR_TO_REPLY) {
	/* Oops, something serious happened, we need to abort */
	return(false);
  }

  /* Now retrigger the watchdog */
  for (retry = 0; retry < WD_RETRIES; retry++) {
	err = do_command(WD_CMD_NOP, 0);
	if (err <= ERR_NONE) {
		/* NOP has been sent successfully */
		return(true);
	}
  }
  return(wd_error(err));
}

